import collections

from pymtl3 import *

#-------------------------------------------------------------------------
# mk_test_case_table
#-------------------------------------------------------------------------

def mk_test_case_table( raw_test_case_table ):

  # First row in test vectors contains port names

  if isinstance(raw_test_case_table[0],str):
    test_param_names = raw_test_case_table[0].split()
  else:
    test_param_names = raw_test_case_table[0]

  TestCase = collections.namedtuple("TestCase",test_param_names)

  ids = []
  test_cases = []
  for row in raw_test_case_table[1:]:
    ids.append( row[0] )
    test_cases.append( TestCase(*row[1:]) )

  return {
    'ids'      : ids,
    'argnames' : ('test_params'),
    'argvalues' : test_cases,
  }

#------------------------------------------------------------------------------
# TestVectorSimulator
#------------------------------------------------------------------------------
# Now we have the same test vector simulator as pymtl2
# https://github.com/cornell-brg/pymtl/blob/master/pclib/test/TestVectorSimulator.py

class TestVectorSimulator:

  def __init__( self, model, test_vectors,
                set_inputs_func, verify_outputs_func, wait_cycles = 0 ):

    self.model               = model
    self.set_inputs_func     = set_inputs_func
    self.verify_outputs_func = verify_outputs_func
    self.test_vectors        = test_vectors
    self.wait_cycles         = wait_cycles

  def run_test( self ):

    self.model.apply( SimulationPass() )

    self.model.sim_reset()

    print()
    for test_vector in self.test_vectors:

      # Set inputs
      self.set_inputs_func( self.model, test_vector )

      self.model.eval_combinational()

      # Print the line trace
      print(self.model.line_trace())

      # Verify outputs
      self.verify_outputs_func( self.model, test_vector )

      self.model.tick()



#=========================================================================
# TranslationImport_adhoc_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Test ad-hoc components with SystemVerilog translation and import."""

from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test import TestVectorSimulator

from .. import TranslationImportPass
from ..translation.behavioral.test.SVBehavioralTranslatorL1_test import (
    test_bit_selection,
    test_comb_assign,
    test_concat,
    test_concat_constants,
    test_concat_mixed,
    test_freevar,
    test_part_selection,
    test_seq_assign,
    test_sext,
    test_unpacked_signal_index,
    test_zext,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL2_test import (
    test_fixed_size_slice,
    test_for_range_lower_upper,
    test_for_range_lower_upper_step,
    test_for_range_upper,
    test_if,
    test_if_bool_op,
    test_if_branches,
    test_if_dangling_else_inner,
    test_if_dangling_else_outter,
    test_if_exp_for,
    test_if_exp_unary_op,
    test_nested_if,
    test_reduce,
    test_tmpvar,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL3_test import (
    test_nested_struct,
    test_packed_array_behavioral,
    test_struct,
    test_struct_const,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL4_test import (
    test_interface,
    test_interface_index,
)
from ..translation.behavioral.test.SVBehavioralTranslatorL5_test import (
    test_subcomponent,
    test_subcomponent_index,
)
from ..translation.structural.test.SVStructuralTranslatorL1_test import (
    test_connect_constant,
    test_port_bit_selection,
    test_port_const,
    test_port_const_array,
    test_port_part_selection,
    test_port_wire,
    test_port_wire_array_index,
)
from ..translation.structural.test.SVStructuralTranslatorL2_test import (
    test_nested_struct_port,
    test_packed_array,
    test_struct_const_structural,
    test_struct_packed_array,
    test_struct_port,
)
from ..translation.structural.test.SVStructuralTranslatorL3_test import (
    test_ifc_decls,
    test_multi_ifc_decls,
)
from ..translation.structural.test.SVStructuralTranslatorL4_test import (
    test_multi_components_ifc_hierarchy_connect,
    test_subcomp_decl,
)


def local_do_test( _m ):
  try:
    _m.elaborate()
    _m.sverilog_translate_import = True
    m = TranslationImportPass()( _m )
    sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
    sim.run_test()
  finally:
    try:
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass



#=========================================================================
# SVTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 15, 2019
"""Provide SystemVerilog translator."""

from pymtl3.passes.backends.generic import RTLIRTranslator

from .behavioral import SVBehavioralTranslator as SV_BTranslator
from .structural import SVStructuralTranslator as SV_STranslator

sverilog_keyword = [
  # Verilog-1995 reserved keywords
  "always", "and", "assign", "begin", "buf", "bufif0", "bufif1", "case",
  "casex", "casez", "cmos", "deassign", "default", "defparam", "disable",
  "edge", "else", "end", "endcase", "endmodule", "endfunction", "endprimitive",
  "endspecify", "endtable", "endtask", "event", "for", "force", "forever",
  "fork", "function", "highz0", "highz1", "if", "ifnone", "initial",
  "inout", "input", "output", "integer", "join", "large", "macromodule",
  "medium", "module", "nand", "negedge", "nmos", "nor", "not", "notif0",
  "notif1", "or", "output", "parameter", "pmos", "posedge", "primitive",
  "pull0", "pull1", "pullup", "pulldown", "rcmos", "real", "realtime",
  "reg", "release", "repeat", "rnmos", "rpmos", "rtran", "rtranif0",
  "rtranif1", "scalared", "small", "specify", "specparam", "strong0",
  "strong1", "supply0", "supply1", "table", "task", "time", "tran",
  "tranif0", "tranif1", "tri", "tri0", "tri1", "triand", "trior",
  "trireg", "vectored", "wait", "wand", "weak0", "weak1", "while",
  "wire", "wor", "xnor", "xor",
  # Verilog-2001 reserved keywords
  "automatic", "cell", "config", "design", "endconfig", "endgenerate",
  "generate", "genvar", "incdir", "include", "instance", "liblist",
  "library", "localparam", "noshowcancelled", "pulsestyle_onevent",
  "pulsestyle_ondetect", "showcancelled", "signed", "unsigned", "use",
  # Verilog-2005 reserved keywords
  "uwire",
  # SystemVerilog-2005 reserved keywords
  "alias", "always_comb", "always_ff", "always_latch", "assert", "assume",
  "before", "bind", "bins", "binsof", "bit", "break", "byte", "chandle",
  "class", "clocking", "const", "constraint", "context", "continue",
  "cover", "covergroup", "coverpoint", "cross", "dist", "do", "endclass",
  "endclocking", "endgroup", "endinterface", "endpackage", "endprimitive",
  "endprogram", "endproperty", "endsequence", "enum", "expect", "export",
  "extends", "extern", "final", "first_match", "foreach", "forkjoin",
  "iff", "ignore_bins", "illegal_bins", "import", "inside", "int", "interface",
  "intersect", "join_any", "join_none", "local", "logic", "longint", "matches",
  "modport", "new", "null", "package", "packed", "priority", "program",
  "property", "protected", "pure", "rand", "randc", "randcase", "randsequence",
  "ref", "return", "sequence", "shortint", "shortreal", "solve", "static",
  "string", "struct", "super", "tagged", "this", "throughout", "timeprecision",
  "timeunit", "type", "typedef", "union", "unique", "var", "virtual", "void",
  "wait_order", "wildcard", "with", "within"
]
sverilog_reserved = set( sverilog_keyword )

def mk_SVTranslator( _RTLIRTranslator, _STranslator, _BTranslator ):

  class _SVTranslator( _RTLIRTranslator, _STranslator, _BTranslator ):

    def get_pretty( s, namespace, attr, newline=True ):
      ret = getattr(namespace, attr, "")
      if newline and (ret and ret[-1] != '\n'):
        ret += "\n"
      return ret

    def is_sverilog_reserved( s, name ):
      return name in sverilog_reserved

    def set_header( s ):
      s.header = \
"""\
//-------------------------------------------------------------------------
// {name}.sv
//-------------------------------------------------------------------------
// This file is generated by PyMTL SystemVerilog translation pass.

"""

    def rtlir_tr_src_layout( s, hierarchy ):
      s.set_header()
      name = s._top_module_full_name
      ret = s.header.format( **locals() )

      # Add struct definitions
      for struct_dtype, tplt in hierarchy.decl_type_struct:
        template = \
"""\
// PyMTL BitStruct {dtype_name} Definition
// At {file_info}
{struct_def}\
"""
        dtype_name = struct_dtype.get_name()
        file_info = struct_dtype.get_file_info()
        struct_def = tplt['def'] + '\n'
        ret += template.format( **locals() )

      # Add component sources
      ret += hierarchy.component_src
      return ret

    def rtlir_tr_components( s, components ):
      return "\n\n".join( components )

    def rtlir_tr_component( s, behavioral, structural ):

      template =\
"""\
// PyMTL Component {component_name} Definition
// {optional_full_name}At {file_info}
module {module_name}
(
{ports});
{body}
endmodule
"""
      component_name = structural.component_name
      file_info = structural.component_file_info
      ports_template = "{port_decls}{ifc_decls}"
      full_name = structural.component_full_name
      module_name = structural.component_unique_name

      if full_name != module_name:
        optional_full_name = f"Full name: {full_name}\n// "
      else:
        optional_full_name = ""

      port_decls = s.get_pretty(structural, 'decl_ports', False)
      ifc_decls = s.get_pretty(structural, 'decl_ifcs', False)
      if port_decls or ifc_decls:
        if port_decls and ifc_decls:
          port_decls += ',\n'
        ifc_decls += '\n'
      ports = ports_template.format(**locals())

      const_decls = s.get_pretty(structural, "decl_consts")
      fvar_decls = s.get_pretty(behavioral, "decl_freevars")
      wire_decls = s.get_pretty(structural, "decl_wires")
      tmpvar_decls = s.get_pretty(behavioral, "decl_tmpvars")
      subcomp_decls = s.get_pretty(structural, "decl_subcomps")
      upblk_decls = s.get_pretty(behavioral, "upblk_decls")
      body = const_decls + fvar_decls + wire_decls + subcomp_decls \
           + tmpvar_decls + upblk_decls
      connections = s.get_pretty(structural, "connections")
      if (body and connections) or (not body and connections):
        connections = '\n' + connections
      body += connections

      s._top_module_name = structural.component_name
      s._top_module_full_name = module_name
      return template.format( **locals() )

  return _SVTranslator

SVTranslator = mk_SVTranslator( RTLIRTranslator, SV_STranslator, SV_BTranslator )



#=========================================================================
# SVStructuralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 3 SystemVerilog structural translator."""

from pymtl3.datatypes import Bits1, Bits32
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..SVStructuralTranslatorL3 import SVStructuralTranslatorL3
from .SVStructuralTranslatorL1_test import check_eq, is_sverilog_reserved


def local_do_test( m ):
  m.elaborate()
  SVStructuralTranslatorL3.is_sverilog_reserved = is_sverilog_reserved
  tr = SVStructuralTranslatorL3( m )
  tr.clear( m )
  tr.translate_structural( m )

  ifcs = tr.structural.decl_ifcs[m]
  conns = tr.structural.connections[m]
  check_eq( ifcs, m._ref_ifcs[m] )
  check_eq( conns, m._ref_conns[m] )

def test_ifc_decls( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc()
      # This 42 will be converted to Bits32(42) by DSL
      connect( s.ifc.msg, 42 )
      # This 1 will be converted to Bits1(1) by DSL
      connect( s.ifc.val, 1 )
  a = A()
  a._ref_ifcs = { a : \
"""\
  output logic [31:0] ifc__msg,
  input logic [0:0] ifc__rdy,
  output logic [0:0] ifc__val\
""" }
  a._ref_conns = { a : \
"""\
  assign ifc__msg = 32'd42;
  assign ifc__val = 1'd1;\
""" }

  # Yosys backend test reference output
  a._ref_ifcs_port_yosys = a._ref_ifcs
  a._ref_ifcs_wire_yosys = { a : "" }
  a._ref_ifcs_conn_yosys = { a : "" }
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.ifc.msg == Bits32(42)
    assert m.ifc.val == Bits1(1)
  a._test_vectors = []
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_multi_ifc_decls( do_test ):
  class InIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class OutIfc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = InIfc()
      s.out = OutIfc()
      connect( s.out, s.in_ )
  a = A()
  a._ref_ifcs = { a : \
"""\
  input logic [31:0] in___msg,
  output logic [0:0] in___rdy,
  input logic [0:0] in___val,
  output logic [31:0] out__msg,
  input logic [0:0] out__rdy,
  output logic [0:0] out__val\
""" }
  a._ref_conns = { a : \
"""\
  assign out__msg = in___msg;
  assign in___rdy = out__rdy;
  assign out__val = in___val;\
""" }

  # Yosys backend test reference output
  a._ref_ifcs_port_yosys = a._ref_ifcs
  a._ref_ifcs_wire_yosys = { a : "" }
  a._ref_ifcs_conn_yosys = { a : "" }
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_.val = Bits1(tv[0])
    m.in_.msg = Bits32(tv[1])
    m.out.rdy = Bits1(tv[2])
  def tv_out( m, tv ):
    assert m.out.val == Bits1(tv[3])
    assert m.out.msg == Bits32(tv[4])
    assert m.in_.rdy == Bits1(tv[5])
  a._test_vectors = [
    [    0,    42,   1,    0,    42,    1 ],
    [    1,    -1,   1,    1,    -1,    1 ],
    [    1,    -2,   0,    1,    -2,    0 ],
    [    0,     2,   0,    0,     2,    0 ],
    [    1,    24,   1,    1,    24,    1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )



#=========================================================================
# SVStructuralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 1 SystemVerilog structural translator."""

from pymtl3.datatypes import Bits1, Bits4, Bits32
from pymtl3.dsl import Component, InPort, OutPort, Wire, connect
from pymtl3.passes.rtlir.util.test_utility import do_test

from ...SVTranslator import sverilog_reserved
from ..SVStructuralTranslatorL1 import SVStructuralTranslatorL1


def is_sverilog_reserved( s, name ):
  return name in sverilog_reserved

def trim( s ):
  string = []
  lines = s.split( '\n' )
  for line in lines:
    _line = line.split()
    _string = "".join( _line )
    if _string and not _string.startswith( '//' ):
      string.append( "".join( line.split() ) )
  return "\n".join( string )

def check_eq( s, t ):
  if isinstance( s, list ) and isinstance( t, list ):
    for _s, _t in zip( s, t ):
      assert trim(_s) == trim(_t)
  else:
    assert trim(s) == trim(t)

def local_do_test( m ):
  m.elaborate()
  SVStructuralTranslatorL1.is_sverilog_reserved = is_sverilog_reserved
  tr = SVStructuralTranslatorL1( m )
  tr.clear( m )
  tr.translate_structural( m )

  ports = tr.structural.decl_ports[m]
  wires = tr.structural.decl_wires[m]
  consts = tr.structural.decl_consts[m]
  conns = tr.structural.connections[m]

  check_eq( ports, m._ref_ports[m] )
  check_eq( wires, m._ref_wires[m] )
  check_eq( consts, m._ref_consts[m] )
  check_eq( conns, m._ref_conns[m] )

def test_port_wire( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.wire_ = Wire( Bits32 )
      s.out = OutPort( Bits32 )
      connect( s.in_, s.wire_ )
      connect( s.wire_, s.out )
  a = A()
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [31:0] out,
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : \
"""\
  logic [31:0] wire_;\
""" }
  a._ref_consts = { a : "" }
  a._ref_conns = { a : \
"""\
  assign wire_ = in_;
  assign out = wire_;\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = a._ref_ports
  a._ref_ports_wire_yosys = { a : "" }
  a._ref_ports_conn_yosys = { a : "" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [    0,   0 ],
    [   42,   42 ],
    [   24,   24 ],
    [   -2,   -2 ],
    [   -1,   -1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_connect_constant( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      connect( 42, s.out )
  a = A()
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : "" }
  a._ref_consts = { a : "" }
  a._ref_conns = { a : \
"""\
  assign out = 32'd42;\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = a._ref_ports
  a._ref_ports_wire_yosys = { a : "" }
  a._ref_ports_conn_yosys = { a : "" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[0])
  a._test_vectors = [
    [    42  ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_port_const( do_test ):
  class A( Component ):
    def construct( s ):
      s.STATE_IDLE = 42
      s.out = OutPort( Bits32 )
      connect( s.STATE_IDLE, s.out )
  a = A()
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : "" }
  a._ref_consts = { a : \
"""\
  localparam [31:0] STATE_IDLE = 32'd42;\
""" }
  # Structural reference to constant number will be replaced
  # with that constant value. This is because the DSL does not
  # add a `my_name` field in _dsl for constant numbers...
  # You can still refer to this constant number in upblks because
  # we define that as a localparam.
  a._ref_conns = { a : \
"""\
  assign out = 32'd42;\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = a._ref_ports
  a._ref_ports_wire_yosys = { a : "" }
  a._ref_ports_conn_yosys = { a : "" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[0])
  a._test_vectors = [
    [    42  ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_port_const_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.STATES = [ 1, 2, 3, 4, 5 ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      for i in range(5):
        connect( s.STATES[i], s.out[i] )
  a = A()
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  output logic [31:0] out [0:4],
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : "" }
  a._ref_consts = { a : \
"""\
  localparam [31:0] STATES [0:4] = '{ 32'd1, 32'd2, 32'd3, 32'd4, 32'd5 };\
""" }
  # Structural reference to constant number will be replaced
  # with that constant value. This is because the DSL does not
  # add a `my_name` field in _dsl for constant numbers...
  # You can still refer to this constant number in upblks because
  # we define that as a localparam ( in the SystemVerilog backend ).
  a._ref_conns = { a : \
"""\
  assign out[0] = 32'd1;
  assign out[1] = 32'd2;
  assign out[2] = 32'd3;
  assign out[3] = 32'd4;
  assign out[4] = 32'd5;\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = { a : \
"""\
  input logic [0:0] clk,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  output logic [31:0] out__2,
  output logic [31:0] out__3,
  output logic [31:0] out__4,
  input logic [0:0] reset\
""" }
  a._ref_ports_wire_yosys = { a : \
"""\
  logic [31:0] out [0:4];\
""" }
  a._ref_ports_conn_yosys = { a : \
"""\
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out__2 = out[2];
  assign out__3 = out[3];
  assign out__4 = out[4];\
""" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[0])
    assert m.out[1] == Bits32(tv[1])
    assert m.out[2] == Bits32(tv[2])
    assert m.out[3] == Bits32(tv[3])
    assert m.out[4] == Bits32(tv[4])
  a._test_vectors = [
    [ 1, 2, 3, 4, 5 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_port_bit_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits1 )
      connect( s.out, s.in_[2] )
  a = A()
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [0:0] out,
  input logic [0:0] reset\
"""
}
  a._ref_wires = { a : "" }
  a._ref_consts = { a : "" }
  # DSL treats a bit selection as a one-bit part selection
  a._ref_conns = { a : \
"""\
  assign out = in_[2:2];\
"""
}

  # Yosys backend test reference output
  a._ref_ports_port_yosys = a._ref_ports
  a._ref_ports_wire_yosys = { a : "" }
  a._ref_ports_conn_yosys = { a : "" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits1(tv[1])
  a._test_vectors = [
    [    0,   0, ],
    [    1,   0, ],
    [    2,   0, ],
    [    3,   0, ],
    [   -1,   1, ],
    [   -2,   1, ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_port_part_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits4 )
      connect( s.out, s.in_[2:6] )
  a = A()
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [3:0] out,
  input logic [0:0] reset\
"""
}
  a._ref_wires = { a : "" }
  a._ref_consts = { a : "" }
  # DSL treats a bit selection as a one-bit part selection
  a._ref_conns = { a : \
"""\
  assign out = in_[5:2];\
"""
}

  # Yosys backend test reference output
  a._ref_ports_port_yosys = a._ref_ports
  a._ref_ports_wire_yosys = { a : "" }
  a._ref_ports_conn_yosys = { a : "" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits4(tv[1])
  a._test_vectors = [
    [    0,   0, ],
    [    1,   0, ],
    [    2,   0, ],
    [    3,   0, ],
    [   -1,  -1, ],
    [   -2,  -1, ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_port_wire_array_index( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      s.wire_ = [ Wire(Bits32) for _ in range(5) ]
      for i in range(5):
        connect( s.wire_[i], s.out[i] )
        connect( s.wire_[i], i )
  a = A()
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  output logic [31:0] out [0:4],
  input logic [0:0] reset\
"""
}
  a._ref_wires = { a : \
"""\
  logic [31:0] wire_ [0:4];\
"""
}
  a._ref_consts = { a : "" }
  # DSL treats a bit selection as a one-bit part selection
  a._ref_conns = { a : \
"""\
  assign out[0] = wire_[0];
  assign wire_[0] = 32'd0;
  assign out[1] = wire_[1];
  assign wire_[1] = 32'd1;
  assign out[2] = wire_[2];
  assign wire_[2] = 32'd2;
  assign out[3] = wire_[3];
  assign wire_[3] = 32'd3;
  assign out[4] = wire_[4];
  assign wire_[4] = 32'd4;\
"""
}

  # Yosys backend test reference output
  a._ref_ports_port_yosys = { a : \
"""\
  input logic [0:0] clk,
  output logic [31:0] out__0,
  output logic [31:0] out__1,
  output logic [31:0] out__2,
  output logic [31:0] out__3,
  output logic [31:0] out__4,
  input logic [0:0] reset\
"""
}
  a._ref_ports_wire_yosys = { a : \
"""\
  logic [31:0] out [0:4];\
"""
}
  a._ref_ports_conn_yosys = { a : \
"""\
  assign out__0 = out[0];
  assign out__1 = out[1];
  assign out__2 = out[2];
  assign out__3 = out[3];
  assign out__4 = out[4];\
"""
}
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[0])
    assert m.out[1] == Bits32(tv[1])
    assert m.out[2] == Bits32(tv[2])
    assert m.out[3] == Bits32(tv[3])
    assert m.out[4] == Bits32(tv[4])
  a._test_vectors = [
    [    0,    1,   2,   3,   4 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )



#=========================================================================
# SVBehavioralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

from pymtl3.datatypes import Bits1, Bits32, Bits64, concat, sext, zext
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure

from ....errors import SVerilogTranslationError
from ...SVTranslator import sverilog_reserved
from ..SVBehavioralTranslatorL1 import BehavioralRTLIRToSVVisitorL1


def is_sverilog_reserved( name ):
  return name in sverilog_reserved

def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = BehavioralRTLIRToSVVisitorL1(is_sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src == m._ref_upblk_srcs[blk.__name__]

def test_comb_assign( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [    0,   0  ],
    [   42,   42 ],
    [   24,   24 ],
    [   -2,   -2 ],
    [   -1,   -1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_seq_assign( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update_ff
      def upblk():
        s.out <<= s.in_
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_ff @(posedge clk) begin : upblk
  out <= in_;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    if tv[1] != '*':
      assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [    0,   '*' ],
    [   42,    0  ],
    [   24,    42 ],
    [   -2,    24 ],
    [   -1,    -2 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_concat( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_1, s.in_2 )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_1, in_2 };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_1 = Bits32(tv[0])
    m.in_2 = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out == Bits64(tv[2])
  a._test_vectors = [
    [    0,    0,     concat(    Bits32(0),    Bits32(0) ) ],
    [   42,    0,     concat(   Bits32(42),    Bits32(0) ) ],
    [   42,   42,     concat(   Bits32(42),    Bits32(42) ) ],
    [   -1,   42,     concat(   Bits32(-1),    Bits32(42) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_concat_constants( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( Bits32(42), Bits32(0) )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { 32'd42, 32'd0 };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out == Bits64(tv[0])
  a._test_vectors = [
    [    concat(    Bits32(42),    Bits32(0) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_concat_mixed( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_, Bits32(0) )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_, 32'd0 };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits64(tv[1])
  a._test_vectors = [
    [  42,  concat(    Bits32(42),    Bits32(0) ) ],
    [  -1,  concat(    Bits32(-1),    Bits32(0) ) ],
    [  -2,  concat(    Bits32(-2),    Bits32(0) ) ],
    [   2,  concat(     Bits32(2),    Bits32(0) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_sext( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = sext( s.in_, 64 )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { { 32 { in_[31] } }, in_ };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits64(tv[0])
  a._test_vectors = [
    [  42,   sext(    Bits32(42),    64 ) ],
    [  -2,   sext(    Bits32(-2),    64 ) ],
    [  -1,   sext(    Bits32(-1),    64 ) ],
    [   2,   sext(     Bits32(2),    64 ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_zext( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = zext( s.in_, 64 )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { { 32 { 1'b0 } }, in_ };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits64(tv[1])
  a._test_vectors = [
    [  42,  zext(    Bits32(42),    64 ) ],
    [  -1,  zext(    Bits32(-1),    64 ) ],
    [  -2,  zext(    Bits32(-2),    64 ) ],
    [   2,  zext(     Bits32(2),    64 ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_sub_component_attr( do_test ):
  class B( Component ):
    def construct( s ):
      s.out_b = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out_b = Bit32(0)
  class A( Component ):
    def construct( s ):
      s.b = B()
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = zext( s.b.out_b, 64 )
  with expected_failure( SVerilogTranslationError, "sub-components" ):
    do_test( A() )

def test_freevar( do_test ):
  class A( Component ):
    def construct( s ):
      STATE_IDLE = Bits32(42)
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_, STATE_IDLE )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_, __const__STATE_IDLE };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits64(tv[1])
  a._test_vectors = [
    [  42,  concat(    Bits32(42),    Bits32(42) ) ],
    [  -1,  concat(    Bits32(-1),    Bits32(42) ) ],
    [  -2,  concat(    Bits32(-2),    Bits32(42) ) ],
    [   2,  concat(     Bits32(2),    Bits32(42) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_, 32'd42 };
end\
"""
}
  do_test( a )

def test_unpacked_signal_index( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(2) ]
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_[0], s.in_[1] )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_[0], in_[1] };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0] = Bits32(tv[0])
    m.in_[1] = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out == Bits64(tv[2])
  a._test_vectors = [
    [  42,   2,  concat(    Bits32(42),     Bits32(2) ) ],
    [  -1,  42,  concat(    Bits32(-1),    Bits32(42) ) ],
    [  -2,  -1,  concat(    Bits32(-2),    Bits32(-1) ) ],
    [   2,  -2,  concat(     Bits32(2),    Bits32(-2) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_bit_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[1]
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_[1];
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits1(tv[1])
  a._test_vectors = [
    [   0,   0 ],
    [  -1,   1 ],
    [  -2,   1 ],
    [   2,   1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_part_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits64 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_[4:36]
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_[35:4];
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = Bits64(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [   -1,   -1 ],
    [   -2,   -1 ],
    [   -4,   -1 ],
    [   -8,   -1 ],
    [  -16,   -1 ],
    [  -32,   -2 ],
    [  -64,   -4 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_sverilog_reserved_keyword( do_test ):
  class A( Component ):
    def construct( s ):
      s.buf = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.buf
  with expected_failure( SVerilogTranslationError, "reserved keyword" ):
    do_test( A() )



#=========================================================================
# StructuralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 3 structural translators."""

import pytest

from pymtl3.datatypes import Bits1, Bits32
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure

from ..StructuralTranslatorL4 import StructuralTranslatorL4
from .TestStructuralTranslator import mk_TestStructuralTranslator


def local_do_test( m ):
  tr = mk_TestStructuralTranslator(StructuralTranslatorL4)(m)
  tr.clear( m )
  tr.translate_structural(m)
  for comp in m._ref_comps.keys():
    decl_comp = tr.structural.decl_subcomps[comp]
    assert decl_comp == m._ref_comps[comp]
  for comp in m._ref_conns.keys():
    connections = tr.structural.connections[comp]
    assert connections == m._ref_conns[comp]

def test_multi_components( do_test ):
  class B( Component ):
    def construct( s ):
      s.out_b = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out_b = Bits32(0)
  class A( Component ):
    def construct( s ):
      s.out_a = OutPort( Bits32 )
      s.b = B()
      connect( s.b.out_b, s.out_a )
  a = A()
  a.elaborate()
  a._ref_comps = {
    a : \
"""\
component_decls:
  component_decl: b Component B
    component_ports:
      component_port: out_b Port of Vector32
    component_ifcs:
""", a.b : \
"""\
component_decls:
"""}
  a._ref_conns = {
    a : \
"""\
connections:
  connection: SubCompAttr CurCompAttr b out_b -> CurCompAttr out_a
""", a.b : \
"""\
connections:
"""}
  a._ref_src = \
"""\
component B
(
port_decls:
  port_decl: out_b Port of Vector32
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:

endcomponent

component A
(
port_decls:
  port_decl: out_a Port of Vector32
interface_decls:
);
const_decls:
freevars:
wire_decls:
component_decls:
  component_decl: b Component B
    component_ports:
      component_port: out_b Port of Vector32
    component_ifcs:
tmpvars:
upblk_srcs:
connections:
  connection: SubCompAttr CurCompAttr b out_b -> CurCompAttr out_a

endcomponent
"""
  do_test( a )

def test_multi_components_ifc_hierarchy_connect( do_test ):
  class OutIfc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.rdy = InPort( Bits1 )
      s.val = OutPort( Bits1 )
  class B( Component ):
    def construct( s ):
      s.out_b = OutPort( Bits32 )
      s.ifc_b = OutIfc()
      connect( s.out_b, 0 )
      connect( s.ifc_b.msg, 0 )
      connect( s.ifc_b.val, 1 )
  class A( Component ):
    def construct( s ):
      s.out_a = OutPort( Bits32 )
      s.b = B()
      s.ifc_a = OutIfc()
      connect( s.b.out_b, s.out_a )
      connect( s.b.ifc_b, s.ifc_a )
  a = A()
  a.elaborate()
  a._ref_comps = {
    a : \
"""\
component_decls:
  component_decl: b Component B
    component_ports:
      component_port: out_b Port of Vector32
    component_ifcs:
      component_ifc: ifc_b InterfaceView OutIfc
        component_ifc_ports:
          component_ifc_port: msg Port of Vector32
          component_ifc_port: rdy Port of Vector1
          component_ifc_port: val Port of Vector1
""", a.b : \
"""\
component_decls:
"""}
  a._ref_conns = {
    a : \
"""\
connections:
  connection: SubCompAttr CurCompAttr b out_b -> CurCompAttr out_a
  connection: IfcAttr SubCompAttr CurCompAttr b ifc_b msg -> IfcAttr CurCompAttr ifc_a msg
  connection: IfcAttr CurCompAttr ifc_a rdy -> IfcAttr SubCompAttr CurCompAttr b ifc_b rdy
  connection: IfcAttr SubCompAttr CurCompAttr b ifc_b val -> IfcAttr CurCompAttr ifc_a val
""", a.b : \
"""\
connections:
  connection: Bits32(0) -> CurCompAttr out_b
  connection: Bits32(0) -> IfcAttr CurCompAttr ifc_b msg
  connection: Bits1(1) -> IfcAttr CurCompAttr ifc_b val
"""}
  a._ref_src = \
"""\
component B
(
port_decls:
  port_decl: out_b Port of Vector32
interface_decls:
  interface_decl: ifc_b InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: Bits32(0) -> CurCompAttr out_b
  connection: Bits32(0) -> IfcAttr CurCompAttr ifc_b msg
  connection: Bits1(1) -> IfcAttr CurCompAttr ifc_b val

endcomponent

component A
(
port_decls:
  port_decl: out_a Port of Vector32
interface_decls:
  interface_decl: ifc_a InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
);
const_decls:
freevars:
wire_decls:
component_decls:
  component_decl: b Component B
    component_ports:
      component_port: out_b Port of Vector32
    component_ifcs:
      component_ifc: ifc_b InterfaceView OutIfc
        component_ifc_ports:
          component_ifc_port: msg Port of Vector32
          component_ifc_port: rdy Port of Vector1
          component_ifc_port: val Port of Vector1
tmpvars:
upblk_srcs:
connections:
  connection: SubCompAttr CurCompAttr b out_b -> CurCompAttr out_a
  connection: IfcAttr SubCompAttr CurCompAttr b ifc_b msg -> IfcAttr CurCompAttr ifc_a msg
  connection: IfcAttr CurCompAttr ifc_a rdy -> IfcAttr SubCompAttr CurCompAttr b ifc_b rdy
  connection: IfcAttr SubCompAttr CurCompAttr b ifc_b val -> IfcAttr CurCompAttr ifc_a val

endcomponent
"""
  do_test( a )

__all__ = [s for s in dir() if s.startswith('test_')]



#=========================================================================
# YosysTranslator_L3_cases_test.py
#=========================================================================
"""Test the yosys-SystemVerilog translator."""

import pytest

from pymtl3.passes.backends.sverilog.translation.structural.test.SVStructuralTranslatorL1_test import (
    check_eq,
)
from pymtl3.passes.backends.sverilog.translation.test.SVTranslator_L3_cases_test import (
    test_ifc_decls,
    test_interface,
    test_interface_index,
    test_multi_ifc_decls,
)
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..YosysTranslator import YosysTranslator


def trim( src ):
  lines = src.split( "\n" )
  ret = []
  for line in lines:
    if not line.startswith( "//" ):
      ret.append( line )
  return "\n".join( ret )

def local_do_test( m ):
  m.elaborate()
  tr = YosysTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, m._ref_src_yosys )



#=========================================================================
# utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 27, 2019
"""Provide helper methods that might be useful to sverilog passes."""

import os
import shutil
import textwrap
from hashlib import blake2b

from pymtl3.passes.rtlir.util.utility import get_component_full_name


def make_indent( src, nindent ):
  """Add nindent indention to every line in src."""
  indent = '  '
  for idx, s in enumerate( src ):
    src[ idx ] = nindent * indent + s

def get_component_unique_name( c_rtype ):
  full_name = get_component_full_name( c_rtype )

  if len( full_name ) < 64:
    return full_name

  comp_name = c_rtype.get_name()
  param_hash = blake2b(digest_size = 8)
  param_hash.update(full_name[len(comp_name):].encode('ascii'))
  param_name = param_hash.hexdigest()
  return comp_name + "__" + param_name

  def get_string( obj ):
    """Return the string that identifies `obj`"""
    if isinstance(obj, type): return obj.__name__
    return str( obj )

def wrap( s ):
  col = shutil.get_terminal_size().columns
  return "\n".join(sum((textwrap.wrap(line, col) for line in s.split("\n")), []))

def expand( v ):
  return os.path.expanduser(os.path.expandvars(v))

def get_dir( cur_file ):
  return os.path.dirname(os.path.abspath(cur_file))+os.path.sep

def get_file_hash( file_path ):
  with open( file_path, 'r' ) as fd:
    hash_inst = blake2b()
    string = ''.join( fd.readlines() ).encode( 'ascii' )
    hash_inst.update(string)
    return hash_inst.hexdigest()



#=========================================================================
# StructuralRTLIRGenL3Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 19, 2019
"""Test the generation of level 1 structural RTLIR."""

import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits1, Bits32
from pymtl3.passes.rtlir.structural.StructuralRTLIRGenL3Pass import (
    StructuralRTLIRGenL3Pass,
)
from pymtl3.passes.rtlir.structural.StructuralRTLIRSignalExpr import *

from .StructuralRTLIRGenL1Pass_test import gen_connections


def test_L3_ifc_view_attr():
  class Ifc( dsl.Interface ):
    def construct( s ):
      s.msg = dsl.InPort( Bits32 )
  class A( dsl.Component ):
    def construct( s ):
      s.in_ = Ifc()
      s.out = dsl.OutPort( Bits32 )
      dsl.connect( s.out, s.in_.msg )
  a = A()
  a.elaborate()
  a.apply( StructuralRTLIRGenL3Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  assert ns.connections == \
    [(InterfaceAttr(CurCompAttr(comp, 'in_'), 'msg'), CurCompAttr(comp, 'out'))]

def test_L3_ifc_view_index():
  class Ifc( dsl.Interface ):
    def construct( s ):
      s.msg = dsl.InPort( Bits32 )
  class A( dsl.Component ):
    def construct( s ):
      s.in_ = [ Ifc() for _ in range(5) ]
      s.out = dsl.OutPort( Bits32 )
      dsl.connect( s.in_[2].msg, s.out )
  a = A()
  a.elaborate()
  a.apply( StructuralRTLIRGenL3Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  assert ns.connections == \
    [(InterfaceAttr(InterfaceViewIndex(CurCompAttr(comp, 'in_'), 2), 'msg'),
     CurCompAttr(comp, 'out'))]

def test_L3_ifc_view_connection():
  class InIfc( dsl.Interface ):
    def construct( s ):
      s.msg = dsl.InPort( Bits32 )
      s.val = dsl.InPort( Bits1 )
      s.rdy = dsl.OutPort( Bits1 )
  class OutIfc( dsl.Interface ):
    def construct( s ):
      s.msg = dsl.OutPort( Bits32 )
      s.val = dsl.OutPort( Bits1 )
      s.rdy = dsl.InPort( Bits1 )
  class A( dsl.Component ):
    def construct( s ):
      s.in_ = InIfc()
      s.out = OutIfc()
      # This will be automatically extended to connect all signals in
      # this interface!
      dsl.connect( s.out, s.in_ )
  a = A()
  a.elaborate()
  a.apply( StructuralRTLIRGenL3Pass( gen_connections( a ) ) )
  ns = a._pass_structural_rtlir_gen
  comp = CurComp(a, 's')
  ref = \
    [
      (InterfaceAttr(CurCompAttr(comp, 'in_'), 'msg'),
      InterfaceAttr(CurCompAttr(comp, 'out'), 'msg')),
      (InterfaceAttr(CurCompAttr(comp, 'in_'), 'val'),
      InterfaceAttr(CurCompAttr(comp, 'out'), 'val')),
      (InterfaceAttr(CurCompAttr(comp, 'out'), 'rdy'),
      InterfaceAttr(CurCompAttr(comp, 'in_'), 'rdy')),
    ]
  # The order of ports is non-deterministic?
  assert ns.connections[0] in ref
  assert ns.connections[1] in ref
  assert ns.connections[2] in ref



#=========================================================================
# YosysBehavioralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Test the SystemVerilog translator implementation."""

from pymtl3.passes.backends.sverilog.translation.behavioral.test.SVBehavioralTranslatorL1_test import (
    is_sverilog_reserved,
    test_bit_selection,
    test_comb_assign,
    test_concat,
    test_concat_constants,
    test_concat_mixed,
    test_freevar,
    test_part_selection,
    test_seq_assign,
    test_sext,
    test_sub_component_attr,
    test_sverilog_reserved_keyword,
    test_unpacked_signal_index,
    test_zext,
)
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..YosysBehavioralTranslatorL1 import YosysBehavioralRTLIRToSVVisitorL1


def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = YosysBehavioralRTLIRToSVVisitorL1(is_sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src == m._ref_upblk_srcs_yosys[blk.__name__]



#=========================================================================
# SVStructuralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 4 SystemVerilog structural translator."""

from pymtl3.datatypes import Bits1, Bits32
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..SVStructuralTranslatorL4 import SVStructuralTranslatorL4
from .SVStructuralTranslatorL1_test import check_eq, is_sverilog_reserved


def local_do_test( m ):
  m.elaborate()
  SVStructuralTranslatorL4.is_sverilog_reserved = is_sverilog_reserved
  tr = SVStructuralTranslatorL4( m )
  tr.clear( m )
  tr.translate_structural( m )
  subcomps = tr.structural.decl_subcomps
  for comp in m._ref_subcomps.keys():
    check_eq( subcomps[comp], m._ref_subcomps[comp] )

def test_subcomp_decl( do_test ):
  class B( Component ):
    def construct( s ):
      s.out_b = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out_b = Bits32(0)
  class A( Component ):
    def construct( s ):
      s.out_a = OutPort( Bits32 )
      s.b = B()
      connect( s.b.out_b, s.out_a )
  a = A()
  a._ref_subcomps = { a : \
"""\
  logic [0:0] b__clk;
  logic [31:0] b__out_b;
  logic [0:0] b__reset;

  B b
  (
    .clk( b__clk ),
    .out_b( b__out_b ),
    .reset( b__reset )
  );\
"""
}

  a._ref_comps_port_yosys = a._ref_subcomps
  a._ref_comps_wire_yosys = { a : "" }
  a._ref_comps_conn_yosys = { a : "" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out_a == Bits32(tv[0])
  a._test_vectors = [
    [    0   ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_multi_components_ifc_hierarchy_connect( do_test ):
  class OutIfc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.rdy = InPort( Bits1 )
      s.val = OutPort( Bits1 )
  class B( Component ):
    def construct( s ):
      s.out_b = OutPort( Bits32 )
      s.ifc_b = OutIfc()
      connect( s.out_b, 0 )
      connect( s.ifc_b.msg, 0 )
      connect( s.ifc_b.val, 1 )
  class A( Component ):
    def construct( s ):
      s.out_a = OutPort( Bits32 )
      s.b = B()
      s.ifc_a = OutIfc()
      connect( s.b.out_b, s.out_a )
      connect( s.b.ifc_b, s.ifc_a )
  a = A()
  a._ref_subcomps = { a : \
"""\
  logic [0:0] b__clk;
  logic [31:0] b__out_b;
  logic [0:0] b__reset;
  logic [31:0] b__ifc_b__msg;
  logic [0:0] b__ifc_b__rdy;
  logic [0:0] b__ifc_b__val;

  B b
  (
    .clk( b__clk ),
    .out_b( b__out_b ),
    .reset( b__reset ),
    .ifc_b__msg( b__ifc_b__msg ),
    .ifc_b__rdy( b__ifc_b__rdy ),
    .ifc_b__val( b__ifc_b__val )
  );\
"""
}

  a._ref_comps_port_yosys = a._ref_subcomps
  a._ref_comps_wire_yosys = { a : "" }
  a._ref_comps_conn_yosys = { a : "" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out_a == Bits32(0)
    assert m.ifc_a.msg == Bits32(0)
    assert m.ifc_a.val == Bits32(1)
  a._test_vectors = []
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )



#=========================================================================
# SVBehavioralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

from pymtl3.datatypes import Bits32, Bits96, bitstruct, concat
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..SVBehavioralTranslatorL3 import BehavioralRTLIRToSVVisitorL3
from .SVBehavioralTranslatorL1_test import is_sverilog_reserved


def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = BehavioralRTLIRToSVVisitorL3(is_sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src == m._ref_upblk_srcs[blk.__name__]

def test_struct( do_test ):
  @bitstruct
  class B:
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_.foo
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_.foo;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [     B(),   0 ],
    [    B(0),   0 ],
    [   B(-1),  -1 ],
    [   B(42),  42 ],
    [   B(-2),  -2 ],
    [   B(10),  10 ],
    [  B(256), 256 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in___foo;
end\
""" }
  do_test( a )

def test_struct_const( do_test ):
  @bitstruct
  class B:
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = B()
      s.out = OutPort( Bits32 )
      s.out_b = OutPort( B )
      @s.update
      def upblk():
        s.out = s.in_.foo
        s.out_b = s.in_
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in_.foo;
  out_b = in_;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out_b == tv[0]
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [       B(),   0 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = 32'd42;
  out_b = 32'd42;
end\
""" }
  do_test( a )

def test_packed_array_behavioral( do_test ):
  @bitstruct
  class B:
    foo: Bits32
    bar: [ Bits32 ] * 2
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits96 )
      @s.update
      def upblk():
        s.out = concat( s.in_.bar[0], s.in_.bar[1], s.in_.foo )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_.bar[0], in_.bar[1], in_.foo };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out == Bits96(tv[1])
  a._test_vectors = [
    [   B(),                              concat(  Bits32(0),   Bits32(0),   Bits32(0) ) ],
    [   B(0, [ Bits32(0) ,Bits32(0)  ] ), concat(  Bits32(0),   Bits32(0),   Bits32(0) ) ],
    [  B(-1, [ Bits32(-1),Bits32(-1) ] ), concat( Bits32(-1),  Bits32(-1),  Bits32(-1) ) ],
    [  B(-1, [ Bits32(42),Bits32(42) ] ), concat( Bits32(42),  Bits32(42),  Bits32(-1) ) ],
    [  B(42, [ Bits32(42),Bits32(42) ] ), concat( Bits32(42),  Bits32(42),  Bits32(42) ) ],
    [  B(42, [ Bits32(-1),Bits32(-1) ] ), concat( Bits32(-1),  Bits32(-1),  Bits32(42) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in___bar[0], in___bar[1], in___foo };
end\
""" }
  do_test( a )

def test_nested_struct( do_test ):
  @bitstruct
  class C:
    woof: Bits32
  @bitstruct
  class B:
    foo: Bits32
    bar: [Bits32]*2
    c: C
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits96 )
      @s.update
      def upblk():
        s.out = concat( s.in_.bar[0], s.in_.c.woof, s.in_.foo )
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in_.bar[0], in_.c.woof, in_.foo };
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out == Bits96(tv[1])
  a._test_vectors = [
    [   B(),                                     concat(   Bits32(0), Bits32(0),   Bits32(0) ) ],
    [   B(0, [ Bits32(0) , Bits32(0)  ], C(5) ), concat(   Bits32(0), Bits32(5),   Bits32(0) ) ],
    [  B(-1, [ Bits32(-1), Bits32(-2) ], C(6) ), concat(  Bits32(-1), Bits32(6),  Bits32(-1) ) ],
    [  B(-1, [ Bits32(42), Bits32(43) ], C(7) ), concat(  Bits32(42), Bits32(7),  Bits32(-1) ) ],
    [  B(42, [ Bits32(42), Bits32(43) ], C(8) ), concat(  Bits32(42), Bits32(8),  Bits32(42) ) ],
    [  B(42, [ Bits32(-1), Bits32(-2) ], C(9) ), concat(  Bits32(-1), Bits32(9),  Bits32(42) ) ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = { in___bar[0], in___c__woof, in___foo };
end\
""" }
  do_test( a )



#=========================================================================
# YosysTranslator_L1_cases_test.py
#=========================================================================
"""Test the yosys-SystemVerilog translator."""

import pytest

from pymtl3.passes.backends.sverilog.translation.structural.test.SVStructuralTranslatorL1_test import (
    check_eq,
)
from pymtl3.passes.backends.sverilog.translation.test.SVTranslator_L1_cases_test import (
    test_bit_selection,
    test_comb_assign,
    test_concat,
    test_concat_constants,
    test_concat_mixed,
    test_connect_constant,
    test_freevar,
    test_part_selection,
    test_port_bit_selection,
    test_port_const_accessed,
    test_port_const_array,
    test_port_const_unaccessed,
    test_port_part_selection,
    test_port_wire,
    test_port_wire_array_index,
    test_seq_assign,
    test_sext,
    test_unpacked_signal_index,
    test_zext,
)
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..YosysTranslator import YosysTranslator


def local_do_test( m ):
  m.elaborate()
  tr = YosysTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, m._ref_src_yosys )



#=========================================================================
# SVBehavioralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

from pymtl3.datatypes import Bits1, Bits32, concat
from pymtl3.dsl import Component, InPort, Interface, OutPort
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure

from ....errors import SVerilogTranslationError
from ..SVBehavioralTranslatorL4 import BehavioralRTLIRToSVVisitorL4
from .SVBehavioralTranslatorL1_test import is_sverilog_reserved


def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  visitor = BehavioralRTLIRToSVVisitorL4(is_sverilog_reserved)
  upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
  m_all_upblks = m.get_update_blocks()
  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src == m._ref_upblk_srcs[blk.__name__]

def test_interface( do_test ):
  class InIfc( Interface ):
    def construct( s ):
      s.val = InPort( Bits1 )
      s.msg = InPort( Bits32 )
      s.rdy = OutPort( Bits1 )
  class OutIfc( Interface ):
    def construct( s ):
      s.val = OutPort( Bits1 )
      s.msg = OutPort( Bits32 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = InIfc()
      s.out = OutIfc()
      @s.update
      def upblk():
        s.out.val = s.in_.val
        s.out.msg = s.in_.msg
        s.in_.rdy = s.out.rdy
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out__val = in___val;
  out__msg = in___msg;
  in___rdy = out__rdy;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_.val = Bits1(tv[0])
    m.in_.msg = Bits32(tv[1])
    m.out.rdy = Bits1(1)
  def tv_out( m, tv ):
    assert m.in_.rdy == Bits1(1)
    assert m.out.val == Bits1(tv[2])
    assert m.out.msg == Bits32(tv[3])
  a._test_vectors = [
    [   1,      0,   1,      0 ],
    [   0,     42,   0,     42 ],
    [   1,     42,   1,     42 ],
    [   1,     -1,   1,     -1 ],
    [   1,     -2,   1,     -2 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_interface_index( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.foo = InPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.in_ = [ Ifc() for _ in range(2) ]
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_[1].foo
  a = A()
  a._ref_upblk_srcs = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in___1__foo;
end\
""" }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0].foo = Bits32(tv[0])
    m.in_[1].foo = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[2])
  a._test_vectors = [
    [    0,    0,      0 ],
    [    0,   42,     42 ],
    [   24,   42,     42 ],
    [   -2,   -1,     -1 ],
    [   -1,   -2,     -2 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { 'upblk' : \
"""\
always_comb begin : upblk
  out = in___foo[1];
end\
""" }
  do_test( a )

def test_interface_array_non_static_index( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.foo = InPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.in_ = [ Ifc() for _ in range(2) ]
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_[s.in_[0].foo].foo
  with expected_failure( SVerilogTranslationError, "static constant expression" ):
    do_test( A() )



#=========================================================================
# YosysStructuralTranslatorL4_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 13, 2019
"""Test the level 4 yosys-SystemVerilog structural translator."""

from pymtl3.datatypes import Bits1, Bits32, bitstruct
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.backends.sverilog.translation.structural.test.SVStructuralTranslatorL1_test import (
    check_eq,
    is_sverilog_reserved,
)
from pymtl3.passes.backends.sverilog.translation.structural.test.SVStructuralTranslatorL4_test import (
    test_multi_components_ifc_hierarchy_connect,
    test_subcomp_decl,
)
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..YosysStructuralTranslatorL4 import YosysStructuralTranslatorL4


def local_do_test( m ):
  m.elaborate()
  YosysStructuralTranslatorL4.is_sverilog_reserved = is_sverilog_reserved
  tr = YosysStructuralTranslatorL4( m )
  tr.clear( m )
  tr.translate_structural( m )

  comps = tr.structural.decl_subcomps[m]

  check_eq( comps["port_decls"], m._ref_comps_port_yosys[m] )
  check_eq( comps["wire_decls"], m._ref_comps_wire_yosys[m] )
  check_eq( comps["connections"], m._ref_comps_conn_yosys[m] )

def test_comp_array( do_test ):
  class B( Component ):
    def construct( s ):
      s.foo = OutPort( Bits32 )
      connect( s.foo, 0 )
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      s.b = [ B() for _ in range(2) ]
      connect( s.out, s.b[1].foo )
  a = A()
  a._ref_comps_port_yosys = { a : \
"""\
  logic [0:0] b__0__clk;
  logic [31:0] b__0__foo;
  logic [0:0] b__0__reset;

  B b__0
  (
    .clk( b__0__clk ),
    .foo( b__0__foo ),
    .reset( b__0__reset )
  );

  logic [0:0] b__1__clk;
  logic [31:0] b__1__foo;
  logic [0:0] b__1__reset;

  B b__1
  (
    .clk( b__1__clk ),
    .foo( b__1__foo ),
    .reset( b__1__reset )
  );\
""" }
  a._ref_comps_wire_yosys = { a : \
"""\
  logic [0:0] b__clk [0:1];
  logic [31:0] b__foo [0:1];
  logic [0:0] b__reset [0:1];\
""" }
  a._ref_comps_conn_yosys = { a : \
"""\
  assign b__0__clk = b__clk[0];
  assign b__1__clk = b__clk[1];
  assign b__foo[0] = b__0__foo;
  assign b__foo[1] = b__1__foo;
  assign b__0__reset = b__reset[0];
  assign b__1__reset = b__reset[1];\
"""
}
  a._ref_conns_yosys = { a : \
"""\
  assign out = b__foo[1];\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out == Bits32(0)
  a._test_vectors = []
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_comp_array_ifc_array_port_array_packed_array( do_test ):
  @bitstruct
  class bstruct:
    bar: [ Bits32 ]
  class Ifc( Interface ):
    def construct( s ):
      s.foo = [ InPort( bstruct ) for _ in range(1) ]
  class B( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(1) ]
      s.out = OutPort( Bits32 )
      connect( s.out, s.ifc[0].foo[0].bar[0] )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.b = [ B() for _ in range(1) ]
      s.out = OutPort( Bits32 )
      connect( s.in_, s.b[0].ifc[0].foo[0].bar[0] )
      connect( s.out, s.b[0].out )
  a = A()
  a._ref_comps_port_yosys = { a : \
"""\
  logic [0:0] b__0__clk;
  logic [31:0] b__0__out;
  logic [0:0] b__0__reset;
  logic [31:0] b__0__ifc__0__foo__0__bar__0;

  B b__0
  (
    .clk( b__0__clk ),
    .out( b__0__out ),
    .reset( b__0__reset ),
    .ifc__0__foo__0__bar__0( b__0__ifc__0__foo__0__bar__0 )
  );\
""" }
  a._ref_comps_wire_yosys = { a : \
"""\
  logic [0:0] b__clk [0:0];
  logic [31:0] b__out [0:0];
  logic [0:0] b__reset [0:0];
  logic [31:0] b__ifc__foo__bar [0:0][0:0][0:0][0:0];
  logic [31:0] b__ifc__foo [0:0][0:0][0:0];\
""" }
  a._ref_comps_conn_yosys = { a : \
"""\
  assign b__0__clk = b__clk[0];
  assign b__out[0] = b__0__out;
  assign b__0__reset = b__reset[0];
  assign b__0__ifc__0__foo__0__bar__0 = b__ifc__foo__bar[0][0][0][0];
  assign b__0__ifc__0__foo__0__bar__0 = b__ifc__foo[0][0][0][31:0];\
"""
}
  a._ref_conns_yosys = { a : \
"""\
  assign b__ifc__foo__bar[0][0][0][0] = in_;
  assign out = b__out[0];\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.b[0].ifc[0].foo[0].bar[0] = Bits32(tv[0])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [    42,  42 ],
    [    -1,  -1 ],
    [    -2,  -2 ],
    [     2,   2 ],
    [    24,  24 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )



#=========================================================================
# SVBehavioralTranslatorL5_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

from pymtl3.datatypes import Bits1, Bits32, concat
from pymtl3.dsl import Component, InPort, Interface, OutPort
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..SVBehavioralTranslatorL5 import BehavioralRTLIRToSVVisitorL5
from .SVBehavioralTranslatorL1_test import is_sverilog_reserved


def local_do_test( m ):
  visitor = BehavioralRTLIRToSVVisitorL5(is_sverilog_reserved)
  for comp, _all_upblks in m._ref_upblk_srcs.items():
    comp.apply( BehavioralRTLIRGenPass() )
    comp.apply( BehavioralRTLIRTypeCheckPass() )
    upblks = comp._pass_behavioral_rtlir_gen.rtlir_upblks
    m_all_upblks = comp.get_update_blocks()
    for blk in m_all_upblks:
      upblk_src = visitor.enter( blk, upblks[blk] )
      upblk_src = "\n".join( upblk_src )
      assert upblk_src == _all_upblks[blk.__name__]

def test_subcomponent( do_test ):
  class B( Component ):
    def construct( s ):
      s.foo = OutPort( Bits32 )
      @s.update
      def upblk():
        s.foo = Bits32(42)
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      s.b = B()
      @s.update
      def upblk():
        s.out = s.b.foo
  a = A()
  a.elaborate()
  a._ref_upblk_srcs = { a : { 'upblk' : \
"""\
always_comb begin : upblk
  out = b__foo;
end\
""" }, a.b : { 'upblk' : \
"""\
always_comb begin : upblk
  foo = 32'd42;
end\
"""
} }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[0])
  a._test_vectors = [
    [    42 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = a._ref_upblk_srcs
  do_test( a )

def test_subcomponent_index( do_test ):
  class B( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.comp = [ B() for _ in range(4) ]
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.comp[2].out
  a = A()
  a.elaborate()
  a._ref_upblk_srcs = { a : { 'upblk' : \
"""\
always_comb begin : upblk
  out = comp__2__out;
end\
""" } }
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[0])
  a._test_vectors = [
    [    0 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  a._ref_upblk_srcs_yosys = { a : { 'upblk' : \
"""\
always_comb begin : upblk
  out = comp__out[2];
end\
""" } }
  do_test( a )



#=========================================================================
# RTLIRTranslator_L2_cases_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 23, 2019
"""Test the RTLIR transaltor."""

import pytest

from ..behavioral.test.BehavioralTranslatorL2_test import *
from ..structural.test.StructuralTranslatorL2_test import *
from .TestRTLIRTranslator import TestRTLIRTranslator


def local_do_test( m ):
  if not m._dsl.constructed:
    m.elaborate()
  tr = TestRTLIRTranslator(m)
  tr.translate( m )
  src = tr.hierarchy.src
  try:
    assert src == m._ref_src
  except AttributeError:
    pass



#=========================================================================
# BehavioralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Test the level 3 behavioral translator."""

from pymtl3.datatypes import Bits32, bitstruct
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..BehavioralTranslatorL3 import BehavioralTranslatorL3
from .TestBehavioralTranslator import mk_TestBehavioralTranslator


def local_do_test( m ):
  m.elaborate()
  tr = mk_TestBehavioralTranslator(BehavioralTranslatorL3)(m)
  tr.clear( m )
  tr.translate_behavioral( m )
  upblk_src = tr.behavioral.upblk_srcs[m]
  decl_freevars = tr.behavioral.decl_freevars[m]
  decl_tmpvars = tr.behavioral.decl_tmpvars[m]
  assert upblk_src == m._ref_upblk_repr
  assert decl_freevars == m._ref_freevar_repr
  assert decl_tmpvars == m._ref_tmpvar_repr

def test_tmp_wire_struct( do_test ):
  @bitstruct
  class B:
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        u = s.in_
        s.out = u.foo
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk
"""
  a._ref_freevar_repr = "freevars:\n"
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk of Struct B
"""
  do_test( a )

def test_multi_tmp_wire_struct( do_test ):
  @bitstruct
  class C:
    bar: Bits32
  @bitstruct
  class B:
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_b = InPort( B )
      s.in_c = InPort( C )
      s.out_b = OutPort( Bits32 )
      s.out_c = OutPort( Bits32 )
      @s.update
      def upblk1():
        u = s.in_b
        s.out_b = u.foo
      @s.update
      def upblk2():
        u = s.in_c
        s.out_c = u.bar
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk1
  upblk_src: upblk2
"""
  a._ref_freevar_repr = "freevars:\n"
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk1 of Struct B
  tmpvar: u in upblk2 of Struct C
"""
  do_test( a )



#=========================================================================
# BehavioralRTLIRL1Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the level 1 behavioral RTLIR passes.

The L1 generation, L1 type check, and visualization passes are invoked. The
generation pass results are verified against a reference AST.
"""

import sys
from copy import copy, deepcopy

import pytest

from pymtl3.datatypes import (
    Bits1,
    Bits2,
    Bits4,
    Bits8,
    Bits16,
    Bits32,
    bitstruct,
    concat,
    sext,
    zext,
)
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL1Pass import (
    BehavioralRTLIRGenL1Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL1Pass import (
    BehavioralRTLIRTypeCheckL1Pass,
)
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError, PyMTLTypeError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure

# TODO: fix all tests with this mark, including in other files (grep!).
# (we use the mark instead of allowing failure more generally because this
#  prevents regressions in all the other tests that *pass* on Python 3)
# If a test passes despite the mark, just remove it from that test!
if sys.version_info[0] == 3:
  XFAIL_ON_PY3 = pytest.mark.xfail(strict=True)
else:
  XFAIL_ON_PY3 = lambda f: f


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  m.elaborate()
  m.apply( BehavioralRTLIRGenL1Pass() )
  m.apply( BehavioralRTLIRTypeCheckL1Pass() )

  try:
    ref = m._rtlir_test_ref
    for blk in m.get_update_blocks():
      upblk = m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ]
      assert upblk == ref[ blk.__name__ ]
  except AttributeError:
    pass

#-------------------------------------------------------------------------
# PyMTL type errors
#-------------------------------------------------------------------------

def test_L1_assign_unagreeable( do_test ):
  @bitstruct
  class B:
    foobar: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        s.out = s.in_
  with expected_failure( PyMTLTypeError, "Unagreeable types" ):
    do_test( A() )

def test_L1_concat_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits8 )
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        s.out = concat( s, s.in_ )
  with expected_failure( PyMTLTypeError, "not a signal" ):
    do_test( A() )

def test_L1_zext_variable_nbits( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits8 )
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        s.out = zext( s.in_, s.in_ )
  with expected_failure( PyMTLTypeError, "not a constant number" ):
    do_test( A() )

def test_L1_zext_small_nbits( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits8 )
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        s.out = zext( s.in_, 4 )
  with expected_failure( PyMTLTypeError, "not greater" ):
    do_test( A() )

def test_L1_sext_variable_nbits( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits8 )
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        s.out = sext( s.in_, s.in_ )
  with expected_failure( PyMTLTypeError, "not a constant number" ):
    do_test( A() )

def test_L1_sext_small_nbits( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits8 )
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        s.out = sext( s.in_, 4 )
  with expected_failure( PyMTLTypeError, "not greater" ):
    do_test( A() )

def test_L1_wrong_attribute( do_test ):
  class A( Component ):
    def construct( s ):
      # s.in_ is not recognized by RTLIR and will be dropped
      s.in_ = 'string'
      s.out = OutPort( Bits16 )
      @s.update
      def upblk():
        s.out = s.in_
  with expected_failure( PyMTLTypeError, "does not have attribute in_" ):
    do_test( A() )

def test_L1_unsupported_attr( do_test ):
  class B( Component ):
    def construct( s ):
      s.out = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits4 )
      s.out = [ OutPort( Bits1 ) for _ in range(5) ]
      s.comp_array = [ B() for _ in range(5) ]
      @s.update
      def upblk():
        s.out = s.comp_array[ 0 ].out
  with expected_failure( PyMTLTypeError, "not supported at L1" ):
    do_test( A() )

def test_L1_array_index_out_of_range( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits1 ) for _ in range(4) ]
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[4]
  with expected_failure( PyMTLTypeError, "array index out of range" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_bit_sel_index_out_of_range( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits4 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[4]
  with expected_failure( PyMTLTypeError, "bit selection index out of range" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_index_on_struct( do_test ):
  @bitstruct
  class B:
    foo: [ Bits4 ] * 4
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[16]
  with expected_failure( PyMTLTypeError, "cannot perform index on Struct" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_slice_on_struct( do_test ):
  @bitstruct
  class B:
    foo: [ Bits4 ] * 4
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[0:16]
  with expected_failure( PyMTLTypeError, "cannot perform slicing on type Struct" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_slice_l_ngt_r( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits4 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[4:0]
  with expected_failure( PyMTLTypeError, "must be larger than the lower bound" ):
    do_test( A() )

def test_L1_slice_variable_bound( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits4 )
      s.slice_l = InPort( Bits2 )
      s.slice_r = InPort( Bits2 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[s.slice_l:s.slice_r]
  with expected_failure( PyMTLTypeError, "slice bounds must be constant" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_slice_out_of_range( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits4 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[0:5]
  with expected_failure( PyMTLTypeError, "slice out of width of signal" ):
    do_test( A() )

#-------------------------------------------------------------------------
# PyMTL type errors: unexpected types
#-------------------------------------------------------------------------

def test_L1_assign_lhs_const( do_test ):
  class A( Component ):
    def construct( s ):
      u, s.v = 42, 42
      @s.update
      def upblk():
        s.v = u
  with expected_failure( PyMTLTypeError, "must be a signal" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_L1_assign_lhs_component( do_test ):
  class B( Component ):
    def construct( s ):
      pass
  class A( Component ):
    def construct( s ):
      s.u = B()
      @s.update
      def upblk():
        s.u = 42
  with expected_failure( PyMTLTypeError, "must be a signal" ):
    do_test( A() )

def test_L1_assign_rhs_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s
  with expected_failure( PyMTLTypeError, "signal or const" ):
    do_test( A() )

def test_L1_zext_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = zext( s, 1 )
  with expected_failure( PyMTLTypeError, "only applies to signals" ):
    do_test( A() )

def test_L1_sext_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = sext( s, 1 )
  with expected_failure( PyMTLTypeError, "only applies to signals" ):
    do_test( A() )

def test_L1_size_cast_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = Bits32( s )
  with expected_failure( PyMTLTypeError, "size casting only applies to signals/consts" ):
    do_test( A() )

def test_L1_attr_signal( do_test ):
  @bitstruct
  class B:
    """Struct class used to trigger certain exception.

    Struct does not belong to level 1. It is just used for testing purposes.
    """
    foobar: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_.foobar
  with expected_failure( PyMTLTypeError, "base of an attribute must be a component" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_L1_index_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_[ s ]
  with expected_failure( PyMTLTypeError, "signal or constant expression" ):
    do_test( A() )

def test_L1_index_base_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s[ 1 ]
  with expected_failure( PyMTLTypeError, "base of an index must be an array or signal" ):
    do_test( A() )

def test_L1_slice_lower_component( do_test ):
  class B( Component ):
    def construct( s ):
      pass
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.idx = B()
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = s.in_[ s.idx:4 ]
  with expected_failure( PyMTLTypeError, "constant expression" ):
    do_test( A() )

def test_L1_slice_upper_component( do_test ):
  class B( Component ):
    def construct( s ):
      pass
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.idx = B()
      s.out = OutPort( Bits4 )
      @s.update
      def upblk():
        s.out = s.in_[ 0:s.idx ]
  with expected_failure( PyMTLTypeError, "constant expression" ):
    do_test( A() )

def test_L1_slice_component( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s[ 0:4 ]
  with expected_failure( PyMTLTypeError, "base of a slice must be a signal" ):
    do_test( A() )

#-------------------------------------------------------------------------
# PyMTL syntax errors
#-------------------------------------------------------------------------

def test_L1_upblk_arg( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk( number ):
        u = number
  with expected_failure( PyMTLSyntaxError, "not have arguments" ):
    do_test( A() )

def test_L1_multi_assign( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        u = v = x = y
  with expected_failure( PyMTLSyntaxError, "multiple targets" ):
    do_test( A() )

def test_L1_copy_arg( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        u = copy(42, 10)
  with expected_failure( PyMTLSyntaxError, "takes exactly 1 argument" ):
    do_test( A() )

def test_L1_deepcopy_arg( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        u = deepcopy(42, 10)
  with expected_failure( PyMTLSyntaxError, "takes exactly 1 argument" ):
    do_test( A() )

def test_L1_slice_step( do_test ):
  class A( Component ):
    def construct( s ):
      v = 42
      @s.update
      def upblk():
        u = v[ 0:16:4 ]
  with expected_failure( PyMTLSyntaxError, "Slice with steps" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_L1_illegal_subscript( do_test ):
  class A( Component ):
    def construct( s ):
      v = 42
      @s.update
      def upblk():
        u = v[ 0:8, 16:24 ]
  with expected_failure( PyMTLSyntaxError, "Illegal subscript" ):
    do_test( A() )

def test_L1_closure_freevar( do_test ):
  class B( Component ): pass
  class A( Component ):
    def construct( s ):
      v = B()
      @s.update
      def upblk():
        u = v
  with expected_failure( PyMTLSyntaxError, "not a sub-component" ):
    do_test( A() )

def test_L1_tmpvar( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        u = 42
  with expected_failure( PyMTLSyntaxError, "Temporary variable" ):
    do_test( A() )

#-------------------------------------------------------------------------
# Call-related syntax errors
#-------------------------------------------------------------------------

@XFAIL_ON_PY3
def test_L1_call_star_arg( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        x = x(*x)
  with expected_failure( PyMTLSyntaxError, "star argument" ):
    do_test( A() )

@XFAIL_ON_PY3
def test_L1_call_double_star_arg( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        x = x(**x)
  with expected_failure( PyMTLSyntaxError, "double-star argument" ):
    do_test( A() )

def test_L1_call_keyword_arg( do_test ):
  class A( Component ):
    def construct( s ):
      xx = 42
      @s.update
      def upblk():
        x = x(x=x)
  with expected_failure( PyMTLSyntaxError, "keyword argument" ):
    do_test( A() )

def test_L1_call_name( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        x = copy.copy( x )
  with expected_failure( PyMTLSyntaxError, "called but is not a name" ):
    do_test( A() )

def test_L1_call_not_found( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        t = undefined_func(u)
  with expected_failure( PyMTLSyntaxError, "function is not found" ):
    do_test( A() )

def test_L1_call_Bits_args( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        x = Bits32( 42, 42 )
  with expected_failure( PyMTLSyntaxError, "exactly one argument" ):
    do_test( A() )

def test_L1_call_concat_args( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        x = concat()
  with expected_failure( PyMTLSyntaxError, "at least one argument" ):
    do_test( A() )

def test_L1_call_zext_num_args( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        x = zext( s )
  with expected_failure( PyMTLSyntaxError, "exactly two arguments" ):
    do_test( A() )

def test_L1_call_sext_num_args( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        x = sext( s )
  with expected_failure( PyMTLSyntaxError, "exactly two arguments" ):
    do_test( A() )

def test_L1_call_unrecognized( do_test ):
  class A( Component ):
    def construct( s ):
      def foo(): pass
      @s.update
      def upblk():
        x = foo()
  with expected_failure( PyMTLSyntaxError, "Unrecognized method" ):
    do_test( A() )

def test_L1_call_global_unrecognized( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        t = do_test(u)
  with expected_failure( PyMTLSyntaxError, "Unrecognized method" ):
    do_test( A() )

#-------------------------------------------------------------------------
# PyMTL syntax errors for unsupported Python syntax
#-------------------------------------------------------------------------

def test_Expr( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        42
  with expected_failure( PyMTLSyntaxError, "Stand-alone expression" ):
    do_test( A() )

def test_Lambda( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = lambda: 42
  with expected_failure( PyMTLSyntaxError, "invalid operation: lambda function" ):
    do_test( A() )

def test_Dict( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = { 1:42 }
  with expected_failure( PyMTLSyntaxError, "invalid type: dict" ):
    do_test( A() )

def test_Set( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = { 42 }
  with expected_failure( PyMTLSyntaxError, "invalid type: set" ):
    do_test( A() )

def test_List( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = [ 42 ]
  with expected_failure( PyMTLSyntaxError, "invalid type: list" ):
    do_test( A() )

def test_Tuple( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = ( 42, )
  with expected_failure( PyMTLSyntaxError, "invalid type: tuple" ):
    do_test( A() )

def test_ListComp( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = [ 42 for _ in range(1) ]
  with expected_failure( PyMTLSyntaxError, "invalid operation: list comprehension" ):
    do_test( A() )

def test_SetComp( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = { 42 for _ in range(1) }
  with expected_failure( PyMTLSyntaxError, "invalid operation: set comprehension" ):
    do_test( A() )

def test_DictComp( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = { 1:42 for _ in range(1) }
  with expected_failure( PyMTLSyntaxError, "invalid operation: dict comprehension" ):
    do_test( A() )

def test_GeneratorExp( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = ( 42 for _ in range(1) )
  with expected_failure( PyMTLSyntaxError, "invalid operation: generator expression" ):
    do_test( A() )

def test_Yield( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = yield
  with expected_failure( PyMTLSyntaxError, "invalid operation: yield" ):
    do_test( A() )

@pytest.mark.xfail( reason = "Syntax not compatible with Python 3" )
def test_Repr( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        # Python 2 only: s.out = `42`
        s.out = repr(42)
  with expected_failure( PyMTLSyntaxError, "invalid operation: repr" ):
    do_test( A() )

def test_Str( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = '42'
  with expected_failure( PyMTLSyntaxError, "invalid operation: str" ):
    do_test( A() )

def test_Classdef( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        class c: pass
  with expected_failure( PyMTLSyntaxError, "invalid operation: classdef" ):
    do_test( A() )

def test_Delete( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        del u
  with expected_failure( PyMTLSyntaxError, "invalid operation: delete" ):
    do_test( A() )

def test_With( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        with u: 42
  with expected_failure( PyMTLSyntaxError, "invalid operation: with" ):
    do_test( A() )

def test_Raise( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        raise 42
  with expected_failure( PyMTLSyntaxError, "invalid operation: raise" ):
    do_test( A() )

@XFAIL_ON_PY3
def test_TryExcept( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        try: 42
        except: 42
  with expected_failure( PyMTLSyntaxError, "invalid operation: try-except" ):
    do_test( A() )

@XFAIL_ON_PY3
def test_TryFinally( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        try: 42
        finally: 42
  with expected_failure( PyMTLSyntaxError, "invalid operation: try-finally" ):
    do_test( A() )

def test_Import( do_test ):
  class A( Component ):
    def construct( s ):
      x = 42
      @s.update
      def upblk():
        import x
  with expected_failure( PyMTLSyntaxError, "invalid operation: import" ):
    do_test( A() )

def test_ImportFrom( do_test ):
  class A( Component ):
    def construct( s ):
      x = 42
      @s.update
      def upblk():
        from x import x
  with expected_failure( PyMTLSyntaxError, "invalid operation: import-from" ):
    do_test( A() )

@pytest.mark.xfail( reason = "Syntax not compatible with Python 3" )
def test_Exec( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        # Python 2 only: exec 42
        exec(42)
  with expected_failure( PyMTLSyntaxError, "invalid operation: exec" ):
    do_test( A() )

def test_Global( do_test ):
  class A( Component ):
    def construct( s ):
      u = 42
      @s.update
      def upblk():
        global u
  with expected_failure( PyMTLSyntaxError, "invalid operation: global" ):
    do_test( A() )

def test_Pass( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        pass
  with expected_failure( PyMTLSyntaxError, "invalid operation: pass" ):
    do_test( A() )

def test_While( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        while 42: 42
  with expected_failure( PyMTLSyntaxError, "invalid operation: while" ):
    do_test( A() )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_ExtSlice( do_test ):
  class A( Component ):
    def construct( s ):
      @s.update
      def upblk():
        42[ 1:2:3, 2:4:6 ]
  with expected_failure( PyMTLSyntaxError, "invalid operation: extslice" ):
    do_test( A() )



#=========================================================================
# StructuralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 3 structural translators."""

import pytest

from pymtl3.datatypes import Bits1, Bits32
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure

from ..StructuralTranslatorL3 import StructuralTranslatorL3
from .TestStructuralTranslator import mk_TestStructuralTranslator


def local_do_test( m ):
  m.elaborate()
  tr = mk_TestStructuralTranslator(StructuralTranslatorL3)(m)
  tr.clear( m )
  tr.translate_structural(m)
  try:
    decl_ifcs = tr.structural.decl_ifcs[m]
    assert decl_ifcs == m._ref_ifcs
    connections = tr.structural.connections[m]
    assert connections == m._ref_conns
  except AttributeError:
    pass

def test_ifc_decls( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc()
      @s.update
      def upblk():
        s.ifc.msg = Bits32(42)
        s.ifc.val = Bits1(1)
  a = A()
  a._ref_ifcs = \
"""\
interface_decls:
  interface_decl: ifc InterfaceView Ifc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
"""
  a._ref_conns = "connections:\n"
  a._ref_src = \
"""\
component A
(
port_decls:
interface_decls:
  interface_decl: ifc InterfaceView Ifc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
  upblk_src: upblk
connections:

endcomponent
"""
  do_test( a )

def test_multi_ifc_decls( do_test ):
  class InIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class OutIfc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = InIfc()
      s.out = OutIfc()
      connect( s.out, s.in_ )
  a = A()
  a._ref_ifcs = \
"""\
interface_decls:
  interface_decl: in_ InterfaceView InIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
  interface_decl: out InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
"""
  a._ref_conns = \
"""\
connections:
  connection: IfcAttr CurCompAttr in_ msg -> IfcAttr CurCompAttr out msg
  connection: IfcAttr CurCompAttr out rdy -> IfcAttr CurCompAttr in_ rdy
  connection: IfcAttr CurCompAttr in_ val -> IfcAttr CurCompAttr out val
"""
  a._ref_src = \
"""\
component A
(
port_decls:
interface_decls:
  interface_decl: in_ InterfaceView InIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
  interface_decl: out InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: IfcAttr CurCompAttr in_ msg -> IfcAttr CurCompAttr out msg
  connection: IfcAttr CurCompAttr out rdy -> IfcAttr CurCompAttr in_ rdy
  connection: IfcAttr CurCompAttr in_ val -> IfcAttr CurCompAttr out val

endcomponent
"""
  do_test( a )

def test_ifc_array_idx( do_test ):
  class InIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class OutIfc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = [ InIfc() for _ in range(5) ]
      s.out = [ OutIfc() for _ in range(5) ]
      for i in range(5):
        connect( s.out[i], s.in_[i] )
  a = A()
  a._ref_ifcs = \
"""\
interface_decls:
  interface_decl: in_ Array[5] of InterfaceView InIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
  interface_decl: out Array[5] of InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
"""
  a._ref_conns = \
"""\
connections:
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 msg -> IfcAttr IfcArrayIdx CurCompAttr out 0 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 0 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 0 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 val -> IfcAttr IfcArrayIdx CurCompAttr out 0 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 msg -> IfcAttr IfcArrayIdx CurCompAttr out 1 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 1 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 1 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 val -> IfcAttr IfcArrayIdx CurCompAttr out 1 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 2 msg -> IfcAttr IfcArrayIdx CurCompAttr out 2 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 2 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 2 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 2 val -> IfcAttr IfcArrayIdx CurCompAttr out 2 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 3 msg -> IfcAttr IfcArrayIdx CurCompAttr out 3 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 3 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 3 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 3 val -> IfcAttr IfcArrayIdx CurCompAttr out 3 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 4 msg -> IfcAttr IfcArrayIdx CurCompAttr out 4 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 4 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 4 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 4 val -> IfcAttr IfcArrayIdx CurCompAttr out 4 val
"""
  a._ref_src = \
"""\
component A
(
port_decls:
interface_decls:
  interface_decl: in_ Array[5] of InterfaceView InIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
  interface_decl: out Array[5] of InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
      interface_port: rdy Port of Vector1
      interface_port: val Port of Vector1
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 msg -> IfcAttr IfcArrayIdx CurCompAttr out 0 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 0 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 0 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 val -> IfcAttr IfcArrayIdx CurCompAttr out 0 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 msg -> IfcAttr IfcArrayIdx CurCompAttr out 1 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 1 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 1 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 1 val -> IfcAttr IfcArrayIdx CurCompAttr out 1 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 2 msg -> IfcAttr IfcArrayIdx CurCompAttr out 2 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 2 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 2 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 2 val -> IfcAttr IfcArrayIdx CurCompAttr out 2 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 3 msg -> IfcAttr IfcArrayIdx CurCompAttr out 3 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 3 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 3 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 3 val -> IfcAttr IfcArrayIdx CurCompAttr out 3 val
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 4 msg -> IfcAttr IfcArrayIdx CurCompAttr out 4 msg
  connection: IfcAttr IfcArrayIdx CurCompAttr out 4 rdy -> IfcAttr IfcArrayIdx CurCompAttr in_ 4 rdy
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 4 val -> IfcAttr IfcArrayIdx CurCompAttr out 4 val

endcomponent
"""
  do_test( a )

# TODO: support nested interface
@pytest.mark.xfail( reason = "TOOD: support nested interface" )
def test_nested_ifc( do_test ):
  class ReqIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class MemReqIfc( Interface ):
    def construct( s ):
      s.req_ifc = ReqIfc()
      s.ctrl_foo = InPort( Bits1 )
      s.ctrl_bar = OutPort( Bits1 )
  class RespIfc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class MemRespIfc( Interface ):
    def construct( s ):
      s.resp_ifc = RespIfc()
      s.ctrl_foo = OutPort( Bits1 )
      s.ctrl_bar = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = [ MemReqIfc() for _ in range(5) ]
      s.out = [ MemRespIfc() for _ in range(5) ]
      for i in range(5):
        connect( s.out[i], s.in_[i] )
  a = A()
  a._ref_ifcs = \
"""\
interface_decls:
  interface_decl: in_ Array[5] of InterfaceView MemReqIfc
    interface_ports:
      interface_port: msg Port of Vector32
  interface_decl: out Array[5] of InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
"""
  a._ref_conns = \
"""\
connections:
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 msg -> IfcAttr IfcArrayIdx CurCompAttr out 0 msg
"""
  a._ref_src = \
"""\
component A
(
port_decls:
interface_decls:
  interface_decl: in_ Array[5] of InterfaceView InIfc
    interface_ports:
      interface_port: msg Port of Vector32
  interface_decl: out Array[5] of InterfaceView OutIfc
    interface_ports:
      interface_port: msg Port of Vector32
);
const_decls:
freevars:
wire_decls:
component_decls:
tmpvars:
upblk_srcs:
connections:
  connection: IfcAttr IfcArrayIdx CurCompAttr in_ 0 msg -> IfcAttr IfcArrayIdx CurCompAttr out 0 msg

endcomponent
"""
  do_test( a )

__all__ = [s for s in dir() if s.startswith('test_')]



"""
#=========================================================================
# OpenLoopCLPass.py
#=========================================================================
# Generate a simple schedule (no Mamba techniques here) based on the
# DAG generated by some previous pass.
#
# Author : Shunning Jiang
# Date   : Apr 20, 2019
"""
from pymtl3.dsl import CalleePort, NonBlockingCalleeIfc
from pymtl3.dsl.errors import UpblkCyclicError

from ..BasePass import BasePass, PassMetadata
from ..errors import PassOrderError
from ..sim.SimpleSchedulePass import SimpleSchedulePass
from ..tracing.CLLineTracePass import CLLineTracePass
from ..tracing.CollectSignalPass import CollectSignalPass
from ..tracing.PrintWavePass import PrintWavePass
from ..tracing.VcdGenerationPass import VcdGenerationPass


class OpenLoopCLPass( BasePass ):
  def __call__( self, top ):
    if not hasattr( top._dag, "all_constraints" ):
      raise PassOrderError( "all_constraints" )

    top._sched = PassMetadata()

    self.schedule_with_top_level_callee( top )

  def schedule_with_top_level_callee( self, top ):

    # Construct the graph with top level callee port

    V = top._dag.final_upblks - top.get_all_update_ff()

    # We collect all top level callee ports/nonblocking callee interfaces

    top_level_callee_ports = top.get_all_object_filter(
      lambda x: isinstance(x, CalleePort) and x.get_host_component() is top )

    top_level_nb_ifcs = top.get_all_object_filter(
      lambda x: isinstance(x, NonBlockingCalleeIfc) and x.get_host_component() is top )

    # We still tell the top level
    method_callee_mapping = {}
    method_guard_mapping  = {}

    # First deal with normal calleeports. We map the actual method to the
    # callee port, and add the port to the vertex set
    for x in top_level_callee_ports:
      if not x.in_non_blocking_interface(): # Normal callee port
        V.add(x)
        assert x.method not in method_callee_mapping
        method_callee_mapping[x.method] = x

    # Then deal with non-blocking callee interfaces. Map the method of the
    # interface to the actual method and set up method-rdy mapping

    for x in top_level_nb_ifcs:
      V.add( x.method )
      method_guard_mapping [x.method] = x.rdy
      assert x.method.method not in method_callee_mapping
      method_callee_mapping[x.method.method] = x.method

    E   = top._dag.all_constraints
    Es  = { v: [] for v in V }
    InD = { v: 0  for v in V }

    for (u, v) in E: # u -> v
      InD[v] += 1
      Es [u].append( v )

    # In addition to existing constraints, we process the constraints that
    # involve top level callee ports. NOTE THAT we assume the user never
    # set the constraint on the actual method object inside the CalleePort
    # In GenDAGPass we already collect those constraints between update
    # blocks and ACTUAL METHODs. We use the ACTUAL METHOD to callee
    # mapping we set up above to avoid missing constraints.

    for (xx, yy) in top._dag.top_level_callee_constraints:

      if xx in method_callee_mapping:
        xx = method_callee_mapping[ xx ]

      if yy in method_callee_mapping:
        yy = method_callee_mapping[ yy ]

      InD[yy] += 1
      Es [xx].append( yy )
      E.add( (xx, yy) )

    # Perform topological sort for a serial schedule.

    schedule = []

    Q = [ v for v in V if not InD[v] ]

    while Q:
      import random
      random.shuffle(Q)

      # Prioritize update blocks instead of method
      # TODO make it O(logn) by balanced BST if needed ...

      u = None
      for i in range(len(Q)):
        if Q[i] not in method_guard_mapping:
          u = Q.pop(i)
          break

      if u is None:
        u = Q.pop()

      if u in method_guard_mapping:
        schedule.append( method_guard_mapping[ u ] )
      schedule.append( u )
      for v in Es[u]:
        InD[v] -= 1
        if not InD[v]:
          Q.append( v )

    # Shunning: we call line trace related pass here.
    CLLineTracePass()( top )
    CollectSignalPass()( top )
    VcdGenerationPass()( top )
    PrintWavePass()( top )
    # Shunning: we reuse ff and posedge schedules from SimpleSchedulePass
    simple = SimpleSchedulePass()
    simple.schedule_ff( top )
    simple.schedule_posedge_flip( top )

    # clear trace before any update block
    schedule.insert( 0, top._tracing.clear_cl_trace )

    # call ff blocks after normal schedule
    schedule.extend( top._sched.schedule_ff )

    # work on tracing
    if hasattr( top._tracing, "vcd_func" ):
      schedule.append( top._tracing.vcd_func )
    if hasattr( top._tracing, "collect_text_sigs" ):
      schedule.append( top._tracing.collect_text_sigs )

    schedule.extend( top._sched.schedule_posedge_flip )

    top._sched.new_schedule_index  = 0
    top._sched.orig_schedule_index = 0

    # Here we are trying to avoid scanning the original schedule that
    # contains methods because we will need isinstance in that case.
    # As a result we created a preprocessed list for execution and use
    # the dictionary to look up the new index of functions.

    # The last element is always line trace
    def print_line_trace():
      print(top.line_trace())

    schedule.append( print_line_trace )

    schedule_no_method = [ x for x in schedule if not isinstance(x, CalleePort) ]
    mapping = { x : i for i, x in enumerate( schedule_no_method ) }

    def wrap_method( top, method,
                     my_idx_new, next_idx_new, schedule_no_method,
                     my_idx_orig, next_idx_orig ):

      def actual_method( *args, **kwargs ):
        i = top._sched.new_schedule_index
        j = top._sched.orig_schedule_index

        if j > my_idx_orig:
          # This means we need to advance the current cycle to the end
          # and then normally execute until we get to the same point.
          # We use original schedule index to handle the case where
          # there are two consecutive methods.
          while i < len(schedule_no_method):
            schedule_no_method[i]()
            i += 1
          i = j = 0
          top.num_cycles_executed += 1

        # We advance from the current point i to the method's position in
        # the schedule without method just to execute those blocks
        while i < my_idx_new:
          schedule_no_method[i]()
          i += 1

        # Execute the method
        ret = method( *args, **kwargs )

        # Execute all update blocks before the next method. Note that if
        # there are several consecutive methods, my_idx_new is equal to next_idx_new
        while i < next_idx_new:
          schedule_no_method[i]()
          i += 1
        j = next_idx_orig

        if i == len(schedule_no_method):
          i = j = 0
          top.num_cycles_executed += 1

        top._sched.new_schedule_index = i
        top._sched.orig_schedule_index = j
        return ret

      return actual_method

    for i, x in enumerate( schedule ):
      if isinstance( x, CalleePort ):
        x.original_method = x.method

        # This is to find the next non-method block's position in the
        # original schedule
        next_func   = i + 1
        while next_func < len(schedule):
          if not isinstance( schedule[next_func], CalleePort ):
            break
          next_func += 1

        # Get the index of the block in the schedule without method
        # This always exists because we append a line trace at the end
        map_next_func = mapping[ schedule[next_func] ]

        # Get the index of the next method in the schedule without method
        next_method = i + 1
        while next_method < len(schedule):
          if isinstance( schedule[next_method], CalleePort ):
            break
          next_method += 1

        # If there is another method after me, I calculate the range of
        # blocks that I need to call and then stop before the user calls
        # the next method.
        if next_method < len(schedule):
          next_func = next_method
          while next_func < len(schedule):
            if not isinstance( schedule[next_func], CalleePort ):
              break
            next_func += 1
          # Get the index in the compacted schedule
          map_next_func_of_next_method = mapping[ schedule[next_func] ]
        else:
          map_next_func_of_next_method = len(schedule_no_method)

        x.method = wrap_method( top, x.method,
                                map_next_func, map_next_func_of_next_method,
                                schedule_no_method,
                                i, next_method )
    top.num_cycles_executed = 0

    return schedule



from .datatypes import *
from .datatypes import _bitwidths
from .dsl import *
from .passes.PassGroups import SimulationPass

__version__ = "0.5.3"

__all__ = [
  'U','M','RD','WR',
  'Wire', 'InPort', 'OutPort', 'Interface', 'CallerPort', 'CalleePort',
  'connect', 'method_port',
  'non_blocking', 'NonBlockingCalleeIfc', 'NonBlockingCallerIfc', 'blocking',

  'SimulationPass',
  'Component', 'Placeholder',

  'sext', 'zext', 'clog2', 'concat', 'reduce_and', 'reduce_or', 'reduce_xor',
  'mk_bits', 'Bits',
  'mk_bitstruct', 'bitstruct',
  'to_bits', 'get_nbits',
] + [ "Bits{}".format(x) for x in _bitwidths ] \
  + [ "b{}".format(x) for x in _bitwidths ]



#=========================================================================
# YosysStructuralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 13, 2019
"""Test the level 3 yosys-SystemVerilog structural translator."""

from pymtl3.datatypes import Bits1, Bits32, bitstruct
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.backends.sverilog.translation.structural.test.SVStructuralTranslatorL1_test import (
    check_eq,
    is_sverilog_reserved,
)
from pymtl3.passes.backends.sverilog.translation.structural.test.SVStructuralTranslatorL3_test import (
    test_ifc_decls,
    test_multi_ifc_decls,
)
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..YosysStructuralTranslatorL3 import YosysStructuralTranslatorL3


def local_do_test( m ):
  m.elaborate()
  YosysStructuralTranslatorL3.is_sverilog_reserved = is_sverilog_reserved
  tr = YosysStructuralTranslatorL3( m )
  tr.clear( m )
  tr.translate_structural( m )

  ports = tr.structural.decl_ifcs[m]
  conns = tr.structural.connections[m]

  check_eq( ports["port_decls"], m._ref_ifcs_port_yosys[m] )
  check_eq( ports["wire_decls"], m._ref_ifcs_wire_yosys[m] )
  check_eq( ports["connections"], m._ref_ifcs_conn_yosys[m] )
  check_eq( conns, m._ref_conns_yosys[m] )

def test_ifc_array( do_test ):
  class InValRdy( Interface ):
    def construct( s, Type ):
      s.msg = InPort( Type )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class OutValRdy( Interface ):
    def construct( s, Type ):
      s.msg = OutPort( Type )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = [ InValRdy( Bits32 ) for _ in range(2) ]
      s.out = [ OutValRdy( Bits32 ) for _ in range(2) ]
      s.sel = InPort( Bits1 )
      s.out_sel = OutPort( Bits32 )
      connect( s.out[0], s.in_[1] )
      connect( s.out[1], s.in_[0] )
      @s.update
      def upblk():
        s.out_sel = s.in_[ s.sel ].msg
  a = A()
  a._ref_ifcs_port_yosys = { a : \
"""\
  input logic [31:0] in___0__msg,
  input logic [31:0] in___1__msg,
  output logic [0:0] in___0__rdy,
  output logic [0:0] in___1__rdy,
  input logic [0:0] in___0__val,
  input logic [0:0] in___1__val,
  output logic [31:0] out__0__msg,
  output logic [31:0] out__1__msg,
  input logic [0:0] out__0__rdy,
  input logic [0:0] out__1__rdy,
  output logic [0:0] out__0__val,
  output logic [0:0] out__1__val\
""" }
  a._ref_ifcs_wire_yosys = { a : \
"""\
  logic [31:0] in___msg [0:1];
  logic [0:0] in___rdy [0:1];
  logic [0:0] in___val [0:1];
  logic [31:0] out__msg [0:1];
  logic [0:0] out__rdy [0:1];
  logic [0:0] out__val [0:1];\
""" }
  a._ref_ifcs_conn_yosys = { a : \
"""\
  assign in___msg[0] = in___0__msg;
  assign in___msg[1] = in___1__msg;
  assign in___0__rdy = in___rdy[0];
  assign in___1__rdy = in___rdy[1];
  assign in___val[0] = in___0__val;
  assign in___val[1] = in___1__val;
  assign out__0__msg = out__msg[0];
  assign out__1__msg = out__msg[1];
  assign out__rdy[0] = out__0__rdy;
  assign out__rdy[1] = out__1__rdy;
  assign out__0__val = out__val[0];
  assign out__1__val = out__val[1];\
"""
}
  a._ref_conns_yosys = { a : \
"""\
  assign out__msg[0] = in___msg[1];
  assign in___rdy[1] = out__rdy[0];
  assign out__val[0] = in___val[1];
  assign out__msg[1] = in___msg[0];
  assign in___rdy[0] = out__rdy[1];
  assign out__val[1] = in___val[0];\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.sel = Bits1(tv[0])
    m.in_[0].msg = Bits32(tv[1])
    m.in_[1].msg = Bits32(tv[2])
  def tv_out( m, tv ):
    assert m.out[0].msg == Bits32(tv[3])
    assert m.out[1].msg == Bits32(tv[4])
    assert m.out_sel == Bits32(tv[5])
  a._test_vectors = [
    [    0,    42,   1,    1,    42,   42 ],
    [    1,    -1,   1,    1,    -1,    1 ],
    [    1,    -2,   0,    0,    -2,    0 ],
    [    0,     2,   0,    0,     2,    2 ],
    [    1,    24,   1,    1,    24,    1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_struct_ifc( do_test ):
  @bitstruct
  class bstruct:
    bar: Bits32
  class Ifc( Interface ):
    def construct( s ):
      s.foo = InPort( bstruct )
  class A( Component ):
    def construct( s ):
      s.in_ = [ Ifc() for _ in range(2) ]
      s.out = OutPort( Bits32 )
      connect( s.out, s.in_[1].foo.bar )
  a = A()
  a._ref_ifcs_port_yosys = { a : \
"""\
  input logic [31:0] in___0__foo__bar,
  input logic [31:0] in___1__foo__bar\
""" }
  a._ref_ifcs_wire_yosys = { a : \
"""\
  logic [31:0] in___foo__bar [0:1];
  logic [31:0] in___foo [0:1];\
""" }
  a._ref_ifcs_conn_yosys = { a : \
"""\
  assign in___foo__bar[0] = in___0__foo__bar;
  assign in___foo__bar[1] = in___1__foo__bar;
  assign in___foo[0][31:0] = in___0__foo__bar;
  assign in___foo[1][31:0] = in___1__foo__bar;\
"""
}
  a._ref_conns_yosys = { a : \
"""\
  assign out = in___foo__bar[1];\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_[0].foo.bar = Bits32(tv[0])
    m.in_[1].foo.bar = Bits32(tv[1])
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[2])
  a._test_vectors = [
    [    42,   1,     1 ],
    [    -1,   1,     1 ],
    [    -2,  42,    42 ],
    [     2,   0,     0 ],
    [    24,  -1,    -1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )



#=========================================================================
# SVTranslator_L3_cases_test.py
#=========================================================================
"""Test the SystemVerilog translator."""

from pymtl3.datatypes import Bits1, Bits32
from pymtl3.dsl import Component, InPort, Interface, OutPort, connect
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..structural.test.SVStructuralTranslatorL1_test import check_eq
from ..SVTranslator import SVTranslator


def local_do_test( m ):
  m.elaborate()
  tr = SVTranslator( m )
  tr.translate( m )
  check_eq( tr.hierarchy.src, m._ref_src )

#-------------------------------------------------------------------------
# Behavioral
#-------------------------------------------------------------------------

def test_interface( do_test ):
  class InIfc( Interface ):
    def construct( s ):
      s.val = InPort( Bits1 )
      s.msg = InPort( Bits32 )
      s.rdy = OutPort( Bits1 )
  class OutIfc( Interface ):
    def construct( s ):
      s.val = OutPort( Bits1 )
      s.msg = OutPort( Bits32 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = InIfc()
      s.out = OutIfc()
      @s.update
      def upblk():
        s.out.val = s.in_.val
        s.out.msg = s.in_.msg
        s.in_.rdy = s.out.rdy
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [0:0] reset,
  input logic [31:0] in___msg,
  output logic [0:0] in___rdy,
  input logic [0:0] in___val,
  output logic [31:0] out__msg,
  input logic [0:0] out__rdy,
  output logic [0:0] out__val
);

  // @s.update
  // def upblk():
  //   s.out.val = s.in_.val
  //   s.out.msg = s.in_.msg
  //   s.in_.rdy = s.out.rdy

  always_comb begin : upblk
    out__val = in___val;
    out__msg = in___msg;
    in___rdy = out__rdy;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_interface_index( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.in_ = [ Ifc() for _ in range(2) ]
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_[1].foo
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset,
  output logic [31:0] in___0__foo,
  output logic [31:0] in___1__foo
);

  // @s.update
  // def upblk():
  //   s.out = s.in_[1].foo

  always_comb begin : upblk
    out = in___1__foo;
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset,
  output logic [31:0] in___0__foo,
  output logic [31:0] in___1__foo
);
  logic [31:0] in___foo [0:1];

  // @s.update
  // def upblk():
  //   s.out = s.in_[1].foo

  always_comb begin : upblk
    out = in___foo[1];
  end

  assign in___0__foo = in___foo[0];
  assign in___1__foo = in___foo[1];

endmodule
"""
  do_test( a )

#-------------------------------------------------------------------------
# Structural
#-------------------------------------------------------------------------

def test_ifc_decls( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc()
      # This 42 will be converted to Bits32(42) by DSL
      connect( s.ifc.msg, 42 )
      # This 1 will be converted to Bits1(1) by DSL
      connect( s.ifc.val, 1 )
  a = A()
  a._ref_src = \
"""\

module A
(
  input logic [0:0] clk,
  input logic [0:0] reset,
  output logic [31:0] ifc__msg,
  input logic [0:0] ifc__rdy,
  output logic [0:0] ifc__val
);

  assign ifc__msg = 32'd42;
  assign ifc__val = 1'd1;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_multi_ifc_decls( do_test ):
  class InIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class OutIfc( Interface ):
    def construct( s ):
      s.msg = OutPort( Bits32 )
      s.val = OutPort( Bits1 )
      s.rdy = InPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.in_ = InIfc()
      s.out = OutIfc()
      connect( s.out, s.in_ )
  a = A()
  a._ref_src = \
"""\

module A
(
  input logic [0:0] clk,
  input logic [0:0] reset,
  input logic [31:0] in___msg,
  output logic [0:0] in___rdy,
  input logic [0:0] in___val,
  output logic [31:0] out__msg,
  input logic [0:0] out__rdy,
  output logic [0:0] out__val
);

  assign out__msg = in___msg;
  assign in___rdy = out__rdy;
  assign out__val = in___val;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )
